''' asim reaction simulator

classes:
reaction
thermo
cstr
plugflow
'''
from __future__ import division
import scipy as sp
import scipy.integrate as itg
import scipy.optimize as opt
import copy
import math



class reaction(object):
    ''' Base reaction class
        no reactor dynamics or thermo added '''
    def __init__(self):
        self.compoundList = {} # index and list of compounds in eqnList
        self.compoundList_r = {} # reverse version of compoundList
        self.thermoList = {}
        self.baseeqnList = {} # raw reaction (without reactor dynamics)
        self.eqnList = {} # list of actual equations for each compound (rate dynamics get added here)
        self.funcList = {} # list of lambda rate equations referenced in eqnLists
        self.feedConc = {} # CAfeed
        self.feedTemp = 0
        self.initial_conditions = {}
        self.initial_temp = {}
        self.rateEqnList = []
        self.reactionList_str = ""
        self._elementary_reactions = []
        self._dTindex = 0
        self.rate_constants = []
        self.heat_of_reaction = []
        self.delta_t = ""
        self.Cp = []
        self.rho = []
        self.E = [] #added activation energy 02/09/12
        self.m = [] #Reaction order Ca^m
        self.map_start = {}
        self.map_end = {}
        self.results = []

    def add(self, reaction, k, m, E, H, Cp, rho):
        """
        Adds and elementary reaction to the class.  Reaction needs to be
        in the form of A + B -> C or if molar constants are needed:
        2*A + 3*B <-> D

        args:   reaction: string, elementary chemical reaction
                k: array, reaction rate constants
                m: array, order of reaction with respect to each component (reactant and oxidant)
                H: double, heat of reaction (J/mol)
                Cp: double, heat capacity of fluid (J/gK)
                rho: double, density of fluid (g/L)

        eg add('A -> B', [7.2e10], [1], 5e4, 0.239, 1000)
        """

        self._elementary_reactions.append(reaction)
        reaction_str = ""
        for iter in self._elementary_reactions:
            reaction_str += iter + "\n"
        self.rate_constants.append(k)
        self.heat_of_reaction.append(H)
        self.Cp.append(Cp)
        self.rho.append(rho)
        self.E.append(E)    # added activation energy 02/09/12
        self.m.append(m)

        # This code parses the chemical reactions to determine how many compounds are present
        # so the index of the termo equation is known

        mol_list = []
        rate_list = []
        list = {}
        self._dTindex = 0
        for line in self._elementary_reactions:
            for term in line.split(): # Divide reaction up by white space e.g 'ES -> E + P' becomes ['ES', '->', 'E', '+', 'P']
                if term == '+': # check for multiple reactants and products in the reaction eg A + B -> C + D
                    continue
                elif term == "<->": # is reaction reversible
                    isReversible = True
                    isReactant = False
                    continue
                elif term == "->": # otherwise normal reaction
                    isReversible = False
                    isReactant = False
                    continue

                mol_list = term.split("*") # Check if there are any stoic coeffients present (eg 2*CO2)
                if len(mol_list) == 1:
                    mol_list = [1] + mol_list # 'ES' to [1, ES]
                else:
                    mol_list = [float(mol_list[0])] + mol_list[1:]
                mol = mol_list[1]



                # check if compound is present in the list of compounds
                if not mol in list: #list.has_key(mol):
                    list.update({mol : self._dTindex})
                    self._dTindex += 1


        self._parsereactionstring(reaction_str)
        # update equations

    def remove(self, idx):
        ''' Removes reaction idx (integer) from the reaction list'''

        self._elementary_reactions.pop(idx)
        reaction_str = ""
        for iter in self._elementary_reactions:
            reaction_str += iter + "\n"
        self._parsereactionstring(reaction_str)
        # update equations

    def __str__(self):
        ''' Built in method to return a string representation of the reactions'''

        out_str = "asim reaction object:\n\n"

        i=0

        for iter in self._elementary_reactions:
            #out_str += str(i) + ": " + iter + "\n"
            out_str += self.rateEqnList[i] + "\n\n"
            i += 1

        reactions = {}
        if len(self.eqnList) > 0:
            reactions = self.eqnList
        else:
            reactions = self.baseeqnList

        out_str += "xdot map: \n" + str(self.compoundList) + "\n"

        xdot_str = "xdot =([\\\n\t"
        for term in reactions:
            xdot_str += reactions[term] + ",\\\n\t"
        xdot_str = xdot_str[0:-4] + "\\\n\t])"
        out_str += xdot_str
        return out_str

    def _parsereactionstring(self, infile):
        '''
        semi private function which parses the reaction string and updates
        the internal list of equations
        '''

        # Constant stuff + declarations
        self.compoundList = {} # index and list of compounds in eqnList
        eqnList = {} # list of actual equations for each compound

        i = -1
        j = -1

        isReactant = True;
        isReversible = False;
        self.rateEqnList = [] # list of equations for each elementary reaction
        k_count = 0
        rate_list = []

        ### iterate over each line in input file
        for line in infile.splitlines():

            if line == '':
                continue
            elif line[0] == '#':
                continue
            else:
                # local variables
                reactants = []
                products = []
                mol_reac = []
                mol_prod = []
                
                rate_str = ""
                isReactant = True

                ### iterate over each term in the equation
                for term in line.split(): # Divide reaction up by white space e.g 'ES -> E + P' becomes ['ES', '->', 'E', '+', 'P']
                    if term == '+': # check for multiple reactants and products in the reaction eg A + B -> C + D
                        continue
                    elif term == "<->": # is reaction reversible
                        isReversible = True
                        isReactant = False
                        continue
                    elif term == "->": # otherwise normal reaction
                        isReversible = False
                        isReactant = False
                        continue

                    """rate_list = term.split("^") # Check if there are any rate coeffients present (eg CO2^1.8) 
                    if len(rate_list) == 1:
                        rate_list = [1] + rate_list # 'ES' to [1.8, ES]
                    else:
                        rate_list = [int(rate_list[0])] +rate_list[1:]
                    self.m.append(rate_list)"""

                    mol_list = term.split("*") # Check if there are any stoic coeffients present (eg 2*CO2)
                    if len(mol_list) == 1:
                        mol_list = [1] + mol_list # 'ES' to [1, ES]
                    else:
                        mol_list = [float(mol_list[0])] + mol_list[1:]
                    mol = mol_list[1]


                    # check if compound is present in the list of compounds
                    if not mol in self.compoundList: # self.compoundList.has_key(mol):
                        i += 1
                        self.compoundList.update({mol : i}) # mapping between xdot index and species
                        self.baseeqnList.update({self.compoundList[mol] : ""}) # create blank entry into list of equations for this compound to be filled in later

                    if isReactant: # add compound to list of reactants for this equation
                        reactants.append(mol_list)
                        mol_reac.append(mol)
                    else: # otherwise add to list of products
                        products.append(mol_list)
                        mol_prod.append(mol)

                j += 1 # number of reactions in this file

                rxnCompounds = list(set(mol_reac + mol_prod)) #Remove any repeating elements in list of reactants + products

                # calculate dT function index
                # this was calculating the index wrong for A->B->C
                # dTindex = len(self._elementary_reactions)
                #dTindex = len(self.compoundList)
                # if dTindex < 2:
                #    dTindex = 2

                #create lambda function call for this reaction
                rate_str = "rxn_" + str(j) + "("
                for term in rxnCompounds:
                    rate_str += "x[" + str(self.compoundList[term]) + "], "
                rate_str = rate_str[0:-1] + " k, x[" + str(self._dTindex)+ "])"


                eqnList.update({j : rate_str})


                # add lambda function call for each compound in ydot

                for term in reactants:
                    self.baseeqnList[self.compoundList[term[1]]] += " -" + str(term[0]) + "*" + rate_str

                for term in products:
                    self.baseeqnList[self.compoundList[term[1]]] += " +" + str(term[0]) + "*" + rate_str


                # generate anonymous (lambda) function for each elementary reaction
                # write out comment + text version of reaction
                rxn_str = "#" + line + "\n"
                rxn_str += "rxn_" + str(j) + " = lambda "

                # write out variables in lambda prototype
                for term in rxnCompounds:
                    rxn_str += term + ", "
                rxn_str = rxn_str[0:-1] + " k, t : "

                # mulitiply by rate constant
                rxn_str += "k[" + str(k_count) + "]*math.exp(-" + str(self.E[k_count]) + "/(8.314*t))* "  # added activation energy to equation 02/09/12
    

                # multiply by each compound in the reaction
                """
                for term in reactants:
                    rxn_str += term[1] + "**" + str(term[0]) + " * "
                rxn_str = rxn_str[0:-3]
                
                """
                m_count = 0

                for term in reactants:
                    rxn_str += term[1] + "**" + str(self.m[k_count][m_count]) + " * "
                    m_count += 1
                rxn_str = rxn_str[0:-3]

                k_count += 1

                # this is for reversible reactions only
                if isReversible == True:
                    rxn_str += " - k[" + str(k_count) + "]*math.exp(-" + str(self.E[k_count]) + "/(8.314*t))* "  # added activation energy to equation 02/09/12
                   
                    """
                    for term in products:
                        rxn_str += term[1] + "**" + str(term[0]) + " * "
                    rxn_str = rxn_str[0:-3]
                    """
                    for term in products:
                        rxn_str += term[1] + "**" + str(self.m[k_count][m_count]) + " * "
                        m_count += 1
                    rxn_str = rxn_str[0:-3]
                    
                    k_count += 1


                self.rateEqnList.append(rxn_str)
                self.reactionList_str = infile

                # create reverse lookup for each compounds location (used for writing out comments)
                self.compoundList_r = {}
                for term in self.compoundList:
                    self.compoundList_r.update({self.compoundList[term] : term})

        self.delta_t = ""
        for iter in eqnList:
            self.delta_t += "+" + eqnList[iter] +"*(" +str(self.heat_of_reaction[j]) + "/(" + str(self.Cp[j]) + "*" + str(self.rho[j]) + "))"


    def addfromfile(self,file):

        infile = open(file,"r").read()
        reaction._parsereactionstring(infile)
        infile.close()


    def rprint(self):
        return self.compoundList

class cstr(object):
    """Continiously Stired Tank Reactor object

    Implements a model of a CSTR reactor 
    """
    def __init__(self, rxn, name):
        """
        Continiously Stired Tank Reactor

        Creates a CSTR reactor and adds a reaction

        args:
        rxn: Reaction object, elementary chemical reaction(s)
        name: string, text name for this reactor

        eg
        cstr(reaction, "Main Reactor")
        """
        self.__volume = 100
        self.__flowrate = 100
        self.initial_conditions = {} # not part of rxn class
        self.initial_temp = 324
        self.coolant_temp = 305
        self.eqn = {}
        self.input = copy.deepcopy(rxn)
        #self.baseeqnList = copy.deepcopy(self.input.baseeqnList)
        self.Cp = 0.239
        self.rho = 1000
        self.UA = 5e4
        self.name = name
        
        # Add this reactor to unit operation list
        length =  len(self.input.eqnList)
        self.input.map_start.update({name : length})

    @property
    def volume(self):
        return self.__volume

    @volume.setter
    def volume(self, volume):
        assert volume > 0, "Volume must be non-zero and non-negative"
        self.__volume = volume

    @property
    def flowrate(self):
        return self.__flowrate

    @flowrate.setter
    def flowrate(self, flowrate):
        assert flowrate >= 0, "Flowrate must be non-negative"
        self.__flowrate = flowrate

    def __str__(self):
        msg = "CSTR object \n\nInitial conditions:"
        msg += str(self.initial_conditions) + "\n"
        msg += "Volume: " + str(self.__volume) + "\n"
        msg += "Flow rate: " + str(self.__flowrate) + "\n"
        msg += "Initial temperature: " + str(self.initial_temp) + "K \n"
        msg += "Coolant temperature: " + str(self.coolant_temp) + "K \n"
        msg += "\n"

        msg += "Reactor equations: \n"

        xdot_str = "xdot =([\\\n\t"
        for term in self.input.eqnList:
            xdot_str += self.input.eqnList[term] + ",\\\n\t"

        xdot_str = xdot_str[0:-4] + "\\\n\t])"
        msg += xdot_str
        return msg

    def _input(self, rxn):
        self.input = copy.deepcopy(rxn)

    def output(self):

        residencetime = self.__flowrate/self.__volume


        len_ = len(self.input.eqnList)
        t = 0
        if len_ > 0:
            len2_ = len(self.input.baseeqnList)+1
            for line in self.input.baseeqnList.values():
                start = 0
                for pos in line:
                    try:
                        start = self.input.baseeqnList[t].index("[", start)
                        start = start + 1
                        end = self.input.baseeqnList[t].index("]", start)
                        index = int(self.input.baseeqnList[t][start:end])
                        self.input.baseeqnList[t] = self.input.baseeqnList[t][:start] + str((len2_+index)) + self.input.baseeqnList[t][end:]

                    except ValueError:
                        break
                t = t + 1

            start = 0
            for pos in self.input.delta_t:
                try:
                    start = self.input.delta_t.index("[", start)
                    start = start + 1
                    end = self.input.delta_t.index("]", start)
                    index = int(self.input.delta_t[start:end])
                    self.input.delta_t = self.input.delta_t[:start] + str((len2_+index)) + self.input.delta_t[end:]
                except ValueError:
                    break

        ###t=0

        for iter in self.input.compoundList_r:

            self.eqn[iter+len_] = self.input.baseeqnList[iter] + " + "  + str(residencetime) + "*(" + str(self.input.feedConc[self.input.compoundList_r[iter]]) + " - x[" + str(iter+len_) + "])"
            self.input.feedConc[self.input.compoundList_r[iter]] = "x[" + str(iter+len_) + "]"

            # need to test if compounds are already present (they should be)
            self.input.initial_conditions[iter+len_] = self.initial_conditions[self.input.compoundList_r[iter]]


        # add temp change due to mass flow of material in reactor
        iter = len(self.input.baseeqnList)
        #self.eqn[iter+len_] = self.input.baseeqnList[iter] + " + " + str(residencetime) + "*(" + str(self.input.feedTemp) + " - x[" + str(iter+len_) + "]) + (" + str(self.UA) + "/(" + str(self.Cp) + "*" + str(self.rho) + "*" + str(self.__volume) + "))*(" + str(self.coolant_temp) + "-" +  "x[" + str(iter+len_) + "])"

        #latest
        self.eqn[iter+len_] = self.input.delta_t + " + " + str(residencetime) + "*(" + str(self.input.feedTemp) + " - x[" + str(iter+len_) + "]) + (" + str(self.UA) + "/(" + str(self.Cp) + "*" + str(self.rho) + "*" + str(self.__volume) + "))*(" + str(self.coolant_temp) + "-" +  "x[" + str(iter+len_) + "])"
        #self.eqn[iter+len_] = self.input.delta_t + " + " + str(residencetime) + "*(" + str(self.input.feedTemp) + " - x[" + str(iter+len_) + "]) + (" + str(self.UA) + "/(" + str(self.input.Cp[iter]) + "*" + str(self.input.rho[iter]) + "*" + str(self.__volume) + "))*(" + str(self.coolant_temp) + "-" +  "x[" + str(iter+len_) + "])"


        self.input.feedTemp = "x[" + str(iter+len_) + "]"
        self.input.initial_conditions[iter+len_] = self.initial_temp

        ###self.input.initial_temp = self.initial_temp


        """
        for iter in self.input.compoundList_r:
            sigma = str(self.input.rate_constants[iter]) + " * math.exp(" + str(self.E) + "/ (" + str(self.R) + "x[" + str(iter+len_) + "]) * x[" + str(iter+len_) + "]"
            self.eqn[iter+len_len(self.input.compoundList_r)] =
            #self.input.feedTemp[]
        """

        #if isempty == True:
        self.input.eqnList.update(self.eqn)
        self.input.map_end.update({self.name : len(self.input.eqnList)})
        return self.input

class bubblecolumn(object):
    """ Creates a bubble column reactor based on the model by Debellefontaine et al
    by splitting the reactor into 2n cstr's with mass transfer between the phases"""
    def __init(self, rxn):
        self.__length = 10
        self.__diameter = 0.1
        self.__flowrate = 0.0001
        self.__n = 3
        self.initial_temp = 350
        self.coolant_temp = 305
        self.input = copy.deepcopy(rxn)
        self.initial_conditions = {"A": 0.0, "B": 0, "C": 0, "D": 0} # not part of rxn class
        self.Cp = 0.239
        self.rho = 1000
        self.UA = 5e4

    @property
    def n(self):
        return self.__n

    @n.setter
    def volume(self, n):
        assert n > 0, "n must be greater than 0"
        self.__n = n

    @property
    def flowrate(self):
        return self.__flowrate

    @flowrate.setter
    def flowrate(self, flowrate):
        assert flowrate >= 0, "Flowrate must be non-negative"
        self.__flowrate = flowrate

    @property
    def length(self):
        return self.__length

    @length.setter
    def length(self, length):
        assert length >= 0, "Length must be greater than 0"
        self.__length = length

    @property
    def diameter(self):
        return self.__diameter

    @diameter.setter
    def diameter(self, diameter):
        assert diameter >= 0, "Diameter must be greater than 0"
        self.__diameter = diameter

    def output(self):

        # calculate volume of each fluid plug (volume of each reactor slice)
        volume = math.pi*math.pow((self.__diameter/2),2)*(self.__length/self.__n)

        # create n cstr objects for the liquid phase
        cstr_in_l = cstr(self.input)
        cstr_in_l.flowrate = self.__flowrate
        cstr_in_l.volume = volume/2
        cstr_in_l.initial_conditions = self.initial_conditions
        cstr_in_l.initial_temp = self.initial_temp
        cstr_in_l.coolant_temp = self.coolant_temp # was missing! APP 23/04/12

        # create n cstr objects for the gas phase
        gas = self.input
        
        gas.rate_constants = [0]
        cstr_in_g = cstr(self.input)
        cstr_in_g.flowrate = self.__flowrate
        cstr_in_g.volume = volume/2
        cstr_in_g.initial_conditions = self.initial_conditions
        cstr_in_g.initial_temp = self.initial_temp
        cstr_in_g.coolant_temp = self.coolant_temp # was missing! APP 23/04/12

        for i in range(1, self.__n):
            cstr_out = cstr_in.output()
            #cstr_in = cstr(cstr_out)
            cstr_in._input(cstr_out)
            #cstr_in.initial_conditions = {"A": 0.0, "B": 0, "C": 0, "D": 0}

        cstr_out = cstr.output(cstr_in)

        #if isempty == True:
        self.input.eqnList.update(cstr_out.eqnList)
        self.input.initial_conditions.update(cstr_out.initial_conditions)

        return self.input

class plugflow(object):

    def __init__(self, rxn, name):
        """
        Plug Flow Reactor

        Creates a PFR reactor and adds a reaction to the reactor

        This model treats the PFR as n CSTR reactors in series,
        by default n=10

        args:
        rxn: Reaction object, elementary chemical reaction(s)
        name: string, text name for this reactor

        eg
        pfr(reaction, "Main Reactor")
        """
        self.__length = 10
        self.__diameter = 0.1
        self.__flowrate = 0.0001
        self.__n = 10
        self.initial_conditions = {"A": 0.0, "B": 0, "C": 0, "D": 0} # not part of rxn class
        self.initial_temp = 350
        self.coolant_temp = 305
        self.eqn = {}
        self.input = copy.deepcopy(rxn)
        self.Cp = 0.239
        self.rho = 1000
        self.UA = 5e4
        self.name = name
        
        # Add this reactor to unit operation list
        length =  len(self.input.eqnList)
        self.input.map_start.update({name : length})

    @property
    def n(self):
        return self.__n

    @n.setter
    def volume(self, n):
        assert n > 0, "n must be greater than 0"
        self.__n = n

    @property
    def flowrate(self):
        return self.__flowrate

    @flowrate.setter
    def flowrate(self, flowrate):
        assert flowrate >= 0, "Flowrate must be non-negative"
        self.__flowrate = flowrate

    @property
    def length(self):
        return self.__length

    @length.setter
    def length(self, length):
        assert length >= 0, "Length must be greater than 0"
        self.__length = length

    @property
    def diameter(self):
        return self.__diameter

    @diameter.setter
    def diameter(self, diameter):
        assert diameter >= 0, "Diameter must be greater than 0"
        self.__diameter = diameter

    def __str__(self):
        msg = "PFR object \n\nInitial conditions:"
        msg += str(self.initial_conditions) + "\n"
        msg += "Length: " + str(self.__length) + "\n"
        msg += "Diameter:" + str(self.__diameter) + "\n"
        msg += "Flow rate: " + str(self.__flowrate) + "\n"
        msg += "Initial temperature: " + str(self.initial_temp) + "K \n"
        msg += "Coolant temperature: " + str(self.coolant_temp) + "K \n"
        msg += "\n"

        msg += "Reactor equations: \n"

        xdot_str = "xdot =([\\\n\t"
        for term in self.input.eqnList:
            xdot_str += self.input.eqnList[term] + ",\\\n\t"

        xdot_str = xdot_str[0:-4] + "\\\n\t])"
        msg += xdot_str
        return msg

    def _input(self, rxn):
        self.input = copy.deepcopy(rxn)

    def output(self):

        # calculate volume of each fluid plug
        volume = math.pi*math.pow((self.__diameter/2),2)*(self.__length/self.__n)

        # create n cstr objects to approximate the pfr
        cstr_in = cstr(self.input)
        cstr_in.flowrate = self.__flowrate
        cstr_in.volume = volume
        cstr_in.initial_conditions = self.initial_conditions
        cstr_in.initial_temp = self.initial_temp
        cstr_in.coolant_temp = self.coolant_temp # was missing! APP 23/04/12

        for i in range(1, self.__n):
            cstr_out = cstr_in.output()
            #cstr_in = cstr(cstr_out)
            cstr_in._input(cstr_out)
            #cstr_in.initial_conditions = {"A": 0.0, "B": 0, "C": 0, "D": 0}

        cstr_out = cstr.output(cstr_in)

        #if isempty == True:
        self.input.eqnList.update(cstr_out.eqnList)
        self.input.initial_conditions.update(cstr_out.initial_conditions)

        return self.input

def optfun(fun, k_est, h_est, xtrue, x0, t):
    x = itg.odeint(fun,x0,t,args=(k_est,h_est,))
    sse = sum(sum(pow(xtrue-x,2)))
    return sse

def regress(rxn, xtrue, tf):

    if(len(rxn.eqnList) == 0):
        rxn.eqnList = rxn.baseeqnList.copy()

    k = []
    h = []
    for term in rxn.rate_constants:
        for rate in term:
            k.append(rate)
    for term in rxn.heat_of_reaction:
        #for heat in term:
        h.append(term)


    # write out lambda rate functions
    ofstr = ""
    for term in rxn.rateEqnList:
        ofstr += term + "\n\n"

    # write out xdot lambda function
    xdot_str = "xdot = lambda x, t, k, H : sp.array([\\\n"
    for term in rxn.eqnList:
        xdot_str += rxn.eqnList[term] + ",\\\n"

    xdot_str = xdot_str[0:-2] + "\\\n])"

    ofstr += xdot_str

    # write compound + equation info to script
    i = len(rxn.compoundList_r)

    # write out initial concentration conditions to script


    ofstr += "\nx0 = sp.array(["
    j=0;
    for term in rxn.eqnList:
        #ofstr += "\t#" + rxn.compoundList_r[j] + "\n\t"
        ofstr += str(rxn.initial_conditions[term]) + ","
        if j < i-1:
            j += 1
        else:
            j = 0
    ofstr = ofstr[0:-1] + "])"

    code_obj = compile(ofstr,'','exec')
    eval(code_obj,globals())

    t = sp.arange(0,tf,0.01)
    k0 = [1.1e4, 3.1e4, 5.5e10]
    h_est = h
    #x = itg.odeint(xdot, x0, t, args=(k,h,))
    #xest = scipy.optimize.fmin_slsqp(optfun,k0,fprime=None,bounds=bounds)

    bounds = [(1e4,2e4),(3e4,4e4),(5e10,6e10)]

    func = lambda kv: optfun(xdot, kv, h_est, xtrue, x0, t)

    kest = opt.fmin_slsqp(func,k0,fprime=None,bounds=bounds)


    return kest

class results(object):
    """Simulation results container
    Properties: 
    UnitOps - List of simulation unit operation result objects
    Map - Dictionary lookup which can be used to index UnitOps

    UnitOps result objects contain the following:


    eg
    rv = asim.sim(obj)
    print(rv.UnitOps["CSTR"].Conc)

    will display the concentration vector for object CSTR
    """
    def __init__(self):
        self.UnitOps = []
        self.Map = {}
    def __str__(self):
        msg = "Simulation results\n"
        msg += "Total unit operations: " + str(len(self.UnitOps)) + "\nUnit Operations:\n"

        for iter in self.Map:
            msg += "Index " + str(self.Map[iter]) + ":" + iter + "\n"
        return msg
    def add(self, obj, map):
        self.UnitOps.append(obj)
        self.Map.update(map)


class result(object):
    """
    result object contains results for a particular unit operation 

    Methods:
    Conc - array of component concentrations
    Temp - array of temperature
    Map  - concentration compound map
    Name - unit operation name
    """
    def __init__(self, conc, temp, map, name):
        self.Conc = conc
        self.Temp = temp
        self.Map = map
        self.Name = name
    def __str__(self):
        msg = "UnitOp: " + self.Name + "\n"
        msg += "Concentration component map: " + str(self.Map) + "\n"

        msg += str(self.Conc) + "\n"
        msg += "Temperature\n"
        msg += str(self.Temp)
        return msg
    


def sim(rxn, tf):
    ''' Simulate passed rxn object to time tf'''

    #import scipy as sp
    #import scipy.integrate as itg

    # place holder variables to keep lint happy
    #x0 = []
    #xdot = None

    if(len(rxn.eqnList) == 0):
        rxn.eqnList = rxn.baseeqnList.copy()

    k = []
    h = []
    for term in rxn.rate_constants:
        for rate in term:
            k.append(rate)
    for term in rxn.heat_of_reaction:
        #for heat in term:
        h.append(term)


    # write out lambda rate functions
    ofstr = ""
    for term in rxn.rateEqnList:
        ofstr += term + "\n\n"

    # write out xdot lambda function
    xdot_str = "xdot = lambda x, t, k, H : sp.array([\\\n"
    for term in rxn.eqnList:
        xdot_str += rxn.eqnList[term] + ",\\\n"

    xdot_str = xdot_str[0:-2] + "\\\n])"

    ofstr += xdot_str

    # write compound + equation info to script
    i = len(rxn.compoundList_r)

    # write out initial concentration conditions to script


    ofstr += "\nx0 = sp.array(["
    j=0;
    for term in rxn.eqnList:
        #ofstr += "\t#" + rxn.compoundList_r[j] + "\n\t"
        ofstr += str(rxn.initial_conditions[term]) + ","
        if j < i-1:
            j += 1
        else:
            j = 0
    ofstr = ofstr[0:-1] + "])"

    code_obj = compile(ofstr,'','exec')
    eval(code_obj,globals())

    t = sp.arange(0,tf,0.01)

    x = itg.odeint(xdot, x0, t, args=(k,h,))

    r = results()
    
    unitops = len(rxn.map_start)
    compounds = len(rxn.compoundList)

    for iter in rxn.map_start:
        cout = x[:,rxn.map_start[iter]:(rxn.map_end[iter]-1)]
        tout = x[:,rxn.map_end[iter]-1:(rxn.map_end[iter])]
        res = result(cout,tout,rxn.compoundList,iter)
        r.add(res,{iter : rxn.map_start[iter]})


    


    """
    ode15s = itg.ode(xdot)
    ode15s.set_integrator('vode', method='bdf', order=15, nsteps=3000)
    ode15s.set_initial_value(x0, 0)
    ode15s.set_f_params([k, h,])
    x = ode15s.integrate(tf)
    """

    return r

def gen_script(rxn, tf):
    #x=0;
    outfilename = "rxn_.py"
    outfile = open(outfilename, "w")
    ofstr = "#!/usr/bin/python\n\nimport scipy as sp\nimport scipy.integrate as itg\nimport matplotlib as mpl\nimport matplotlib.pyplot as plt\n\n"
    ofstr += "'''\n## Reaction ##\n\n"
    ofstr += rxn.reactionList_str + "\n\n"

    ofstr += "## Mapping ##\n\n"

    #
    if len(rxn.eqnList) == 0:
        rxn.eqnList = rxn.baseeqnList
        rxn.initial_conditions = rxn.feedConc

    # write compound + equation info to script
    i = len(rxn.compoundList_r)
    j=0
    for term in rxn.eqnList:
        ofstr += rxn.compoundList_r[j] + "\t" + str(term) + "\t" + rxn.eqnList[term] + "\n"
        if j < i-1:
            j += 1
        else:
            j = 0
    ofstr += "'''\n\n"

    ofstr += "def main():\n\n"

    ofstr += "\t# Rate constants\n"
    ofstr += "\tk = ["
    for term in rxn.rate_constants:
        for rate in term:
            ofstr += str(rate) + ", "
    ofstr = ofstr[0:-2] + "]\n\n"

    # write out lambda rate functions
    ofstr += "\t# Kinetic rate equations\n"
    for term in rxn.rateEqnList:
        ofstr += "\t" + term + "\n\n"

    # write out xdot lambda function
    xdot_str = "\txdot = lambda x, t, k: sp.array([\\\n\t"
    for term in rxn.eqnList:
        xdot_str += rxn.eqnList[term] + ",\\\n\t"

    xdot_str = xdot_str[0:-4] + "\\\n\t])"

    ofstr += xdot_str

    # write out initial concentration conditions to script
    ofstr += "\t\n\n#Initial concentrations:\n\tx0 = sp.array([\\\n"
    j=0;
    for term in rxn.eqnList:
        ofstr += "\t#" + rxn.compoundList_r[j] + "\n\t"
        ofstr += str(rxn.initial_conditions[term]) + ",\\\n"
        if j < i-1:
            j += 1
        else:
            j = 0
    ofstr += "\t])\n\n"


    ofstr += "\tt = sp.arange(0, " + str(tf) + ", 0.01)\n"

    ofstr += "\tx = sp.integrate.odeint(xdot, x0, t, args=(k,))\n"

    ofstr += "\treturn x\n\n"

#    ofstr += "plt.plot(x)\nplt.draw()\nplt.show()\n"
    ofstr += "if __name__ == '"'__main__'"':\n\tmain()"
    #print ofstr
    outfile.write(ofstr)
    outfile.close()


def param_sim(reaction, data, k0, bounds):
    ''' '''
    # Optimisation function
'''
import rxn_

    def optfun(k_est):
        x = sp.integrate.odeint(rxn_.main(),x0,dt,args=(k_est,))
        sse = sum(sum(pow(xtrue-x,2)))
        return sse

# Initial parameter guess
k0 = array([1,1,1,1])

bounds = [(0,10),(0,10),(0,10),(0,10)]

#xest = scipy.optimize.fmin_l_bfgs_b(optfun,k0,fprime=None,args=(),approx_grad=1,bounds=bounds,pgtol=1e-6)
#print(xest[0])
xest = scipy.optimize.fmin_slsqp(optfun,k0,fprime=None,bounds=bounds)
#xest = scipy.optimize.fmin_tnc(optfun,k0,fprime=None,args=(),approx_grad=1,bounds=bounds,pgtol=1e-6)
'''
